#= require spec_helper


describe 'BrestLoader', ->
  
  beforeEach ->
    define_test_data this
    @brest_loader = new slender_data.BrestLoader @test_params
    @callback = sinon.spy()


  describe 'create', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success', user: @user_1
      @brest_loader.create email: 'one@example.com', @callback

    afterEach ->
      jQuery.ajax.restore()

    it 'makes the ajax request to the REST create endpoint', ->
      expect(jQuery.ajax).to.have.been.calledOnce
      expect(jQuery.ajax).to.have.been.calledWithMatch type: 'POST'
      expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json'
      expect(jQuery.ajax).to.have.been.calledWithMatch
        data: { user: { email: 'one@example.com' }}

    it 'returns the entry from the server as a query', ->
      expect(@callback).to.have.been.calledOnceWith user: @user_1



  describe 'create_many', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success',
                                          users: [@user_1, @user_2]
      @brest_loader.create_many [ {email: 'one@example.com'},
                                  {email: 'two@example.com'} ],
                                @callback

    afterEach ->
      jQuery.ajax.restore()

    it 'makes the ajax request to the bulk create endpoint', ->
      expect(jQuery.ajax).to.have.been.calledOnce
      expect(jQuery.ajax).to.have.been.calledWithMatch type: 'POST'
      expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json'
      expect(jQuery.ajax).to.have.been.calledWithMatch
        data: { users: [ { email: 'one@example.com' },
                         { email: 'two@example.com' } ] }

    it 'returns the entry from the server as a multi_query', ->
      expect(@callback).to.have.been.calledOnceWith users: [@user_1, @user_2]



  describe 'delete', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success'

    afterEach ->
      jQuery.ajax.restore()


    context 'entry query given', ->

      beforeEach ->
        @brest_loader.delete user: {username: 'one'}, @callback

      it 'makes the ajax request to the REST delete endpoint', ->
        expect(jQuery.ajax).to.have.been.calledOnce
        expect(jQuery.ajax).to.have.been.calledWithMatch type: 'DELETE'
        expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users/one.json'

      it 'calls the callback when done', ->
        expect(@callback).to.have.been.calledOnce


    context 'id query given', ->

      beforeEach ->
        @brest_loader.delete username: 'one', @callback

      it 'makes the ajax request', ->
        expect(jQuery.ajax).to.have.been.calledOnce
        expect(jQuery.ajax).to.have.been.calledWithMatch type: 'DELETE'
        expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users/one.json'

      it 'calls the callback when done', ->
        expect(@callback).to.have.been.calledOnce



  describe 'delete_many', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success'

    afterEach ->
      jQuery.ajax.restore()


    context 'entries multi-query given', ->

      beforeEach ->
        @brest_loader.delete_many users: [{username: 'one'}, {username: 'two'}],
                                  @callback

      it 'makes the ajax request to the bulk delete endpoint', ->
        expect(jQuery.ajax).to.have.been.calledOnce
        expect(jQuery.ajax).to.have.been.calledWithMatch type: 'DELETE'
        expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json?usernames=one+two'

      it 'returns the entry from the server', ->
        expect(@callback).to.have.been.calledOnce


    context 'ids multi-query given', ->

      beforeEach ->
        @brest_loader.delete_many usernames: ['one', 'two'], @callback

      it 'makes the ajax request to the bulk delete endpoint', ->
        expect(jQuery.ajax).to.have.been.calledOnce
        expect(jQuery.ajax).to.have.been.calledWithMatch type: 'DELETE'
        expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json?usernames=one+two'

      it 'calls the callback when done', ->
        expect(@callback).to.have.been.calledOnce



  describe 'entry_url', ->

    it 'works without a given base_url value', ->
      @test_params.base_url = undefined
      brest_loader = new slender_data.BrestLoader @test_params
      expect(brest_loader.entry_url id: 3).to.equal '/users/3.json'

    it 'uses a given base url in the generated url', ->
      @test_params.base_url = '/api'
      brest_loader = new slender_data.BrestLoader @test_params
      expect(brest_loader.entry_url id: 3).to.equal '/api/users/3.json'

    it 'works with id queries', ->
      brest_loader = new slender_data.BrestLoader @test_params
      expect(brest_loader.entry_url id: 3).to.equal '/users/3.json'

    it 'works with entry queries', ->
      brest_loader = new slender_data.BrestLoader @test_params
      expect(brest_loader.entry_url user: {username: 'jim'}).to.equal '/users/jim.json'


  describe 'entries_url', ->

    it 'returns the entries url without parameters', ->
      expect(@brest_loader.entries_url()).to.eql '/users.json'

    it 'returns the entries url with the given data', ->
      expect(@brest_loader.entries_url(foo: 'bar')).to.eql '/users.json?foo=bar'


  describe 'merge', ->

    beforeEach ->
      @merged_user = username: 'merged', timestamp: 1
      @server_merge_api = sinon.stub(jQuery, 'ajax')
                               .yieldsTo 'success', user: @merged_user
      @callback = sinon.stub()
      @brest_loader.merge [1, 2], @merged_user, @callback

    afterEach ->
      @server_merge_api.restore()

    it 'makes the server call', ->
      expect(@server_merge_api).to.have.been.calledOnce

    it 'calls the server with the right arguments', ->
      expect(@server_merge_api).to.have.been.calledWithMatch
        data: old_snippet_ids: [1, 2], merged_snippet_data: @merged_user

    it 'calls the given callback with the server response', ->
      expect(@callback).to.have.been.calledOnceWith user: @merged_user


  describe 'read', ->

    beforeEach ->
      sinon.stub(jQuery, 'get').yields user: @user_1

    afterEach ->
      jQuery.get.restore()


    context 'entry query given', ->

      beforeEach ->
        @brest_loader.read user: {username: 'one'}, @callback

      it 'makes an ajax request to the REST read endpoint', ->
        expect(jQuery.get).to.have.been.calledOnceWith '/users/one.json'

      it 'returns the entry from the server as a query', ->
        expect(@callback).to.have.been.calledOnceWith user: @user_1


    context 'id query given', ->

      beforeEach ->
        @brest_loader.read username: 'one', @callback

      it 'makes an ajax request to the REST read endpoint', ->
        expect(jQuery.get).to.have.been.calledOnceWith '/users/one.json'

      it 'returns the entry from the server as a query', ->
        expect(@callback).to.have.been.calledOnceWith user: @user_1


  describe 'read_many', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success',
                                          users: [@user_1, @user_2]

    afterEach ->
      jQuery.ajax.restore()


    context 'entries multi-query given', ->

      beforeEach ->
        @brest_loader.read_many users: [{username: 'one'}, {username: 'two'}],
                                @callback

      it 'makes an ajax request to the bulk read endpoint', ->
        expect(jQuery.ajax).to.have.been.calledOnce
        expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json'
        expect(jQuery.ajax).to.have.been.calledWithMatch data: {usernames: 'one two'}

      it 'returns the entry from the server as a multi_query', ->
        expect(@callback).to.have.been.calledOnceWith users: [@user_1, @user_2]


    context 'ids multi-query given', ->

      beforeEach ->
        @brest_loader.read_many usernames: ['one', 'two', 'three'], @callback

      it 'makes an ajax request to the bulk read endpoint', ->
        expect(jQuery.ajax).to.have.been.calledOnce
        expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json'
        expect(jQuery.ajax).to.have.been.calledWithMatch data: {usernames: 'one two three'}

      it 'returns the entry from the server as a multi_query', ->
        expect(@callback).to.have.been.calledOnceWith users: [@user_1, @user_2]



  describe 'update', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success', user: @user_1

    afterEach ->
      jQuery.ajax.restore()


    beforeEach ->
      @brest_loader.update username: 'one', email: 'new@example.com',
                           @callback

    it 'makes the ajax request to the REST update endpoint', ->
      expect(jQuery.ajax).to.have.been.calledOnce
      expect(jQuery.ajax).to.have.been.calledWithMatch type: 'PUT'
      expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users/one.json'
      expect(jQuery.ajax).to.have.been.calledWithMatch data: { user: { email: 'new@example.com' }}

    it 'returns the entry from the server as a query', ->
      expect(@callback).to.have.been.calledOnceWith user: @user_1



  describe 'update_many', ->

    beforeEach ->
      sinon.stub(jQuery, 'ajax').yieldsTo 'success',
                                          users: [@user_1, @user_2]
      @brest_loader.update_many [ {username: 'one', email: 'one@example.com'},
                                  {username: 'two', email: 'two@example.com'} ],
                                @callback

    afterEach ->
      jQuery.ajax.restore()

    it 'makes the ajax request to the bulk update endpoint', ->
      expect(jQuery.ajax).to.have.been.calledOnce
      expect(jQuery.ajax).to.have.been.calledWithMatch type: 'PUT'
      expect(jQuery.ajax).to.have.been.calledWithMatch url: '/users.json'
      expect(jQuery.ajax).to.have.been.calledWithMatch
        data: { users: [ { username: 'one', email: 'one@example.com' },
                         { username: 'two', email: 'two@example.com' } ] }

    it 'returns the entry from the server as a multi_query', ->
      expect(@callback).to.have.been.calledOnceWith users: [@user_1, @user_2]


  describe 'url_data_encode', ->

    it 'returns a string version of the given key/value data', ->
      expect(@brest_loader.url_data_encode(id: 'one')).to.equal "id=one"

    it 'returns a string version of the given key/array data', ->
      expect(@brest_loader.url_data_encode(id: ['one', 'two'])).to.equal "id=one+two"

    it 'concatenates multiple elements with &', ->
      expect(@brest_loader.url_data_encode(foo: 'bar', id: ['one', 'two'])).to.equal "foo=bar&id=one+two"

