#= require spec_helper
#= require slender_data/entry_store


describe 'EntryStore', ->

  beforeEach ->
    define_test_data this
    @entry_store = new slender_data.EntryStore @test_params


  describe 'add', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @entry_store.add @user_1

      it 'adds the given entry to the cache', ->
        expect(@entry_store.get user: @user_1).to.eql @user_1

      it 'also updates an existing client_data entry', ->
        @client_entry = @entry_store.get user: @user_1
        @client_entry.timestamp = 99
        server_entry = _.clone @user_1
        server_entry.email = "server_two@example.com"
        result = @entry_store.add server_entry
        expect(@entry_store.get username: 'one').to.eql server_entry

      it 'returns nothing', ->
        expect(@result).to.be.undefined


    describe 'partial entries management', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @entry_store = new slender_data.EntryStore @test_params

      context 'partial entry cached', ->

        beforeEach ->
          @entry_store.add @partial_user_1

        it 'replaces the cached entry with the given full entry', ->
          @entry_store.add @full_user_1
          expect(@entry_store.get username: 'one').to.eql @full_user_1

      context 'full entry cached', ->

        beforeEach ->
          @entry_store.add @full_user_1

        it 'does not replace the cached value with the given partial entry', ->
          @entry_store.add @partial_user_1
          expect(@entry_store.get username: 'one').to.eql @full_user_1

    describe 'return-editables option', ->

      it 'returns editable versions of the option is set', ->
        result = @entry_store.add @user_1, yes
        expect(result).to.be.an.editable_version_of @user_1

      it 'returns nothing if the option is not set', ->
        result = @entry_store.add @user_1, no
        expect(result).to.be.undefined


  describe 'add_many', ->

    describe 'default behavior', ->

      beforeEach ->
        @result = @entry_store.add_many [@user_1, @user_2]

      it 'adds the given entries to the cache', ->
        expect(@entry_store.get user: @user_1).to.eql @user_1
        expect(@entry_store.get user: @user_2).to.eql @user_2

      it 'returns nothing', ->
        expect(@result).to.be.undefined


    describe 'return_editable_version option', ->

      it 'returns editable versions of the given data if set', ->
        result = @entry_store.add_many [@user_1, @user_2], yes
        expect(result[0]).to.be.an.editable_version_of @user_1
        expect(result[1]).to.be.an.editable_version_of @user_2

      it 'returns nothing if not set', ->
        expect(@entry_store.add_many [@user_1, @user_2], no).to.be.undefined


  describe 'delete', ->

    beforeEach ->
      @entry_store.add_many [@user_1, @user_2]


    context 'entry query given', ->

      beforeEach ->
        @entry_store.delete user: @user_1

      it 'removes the entry described by the given query from this store', ->
        expect(@entry_store.get user: @user_1).to.be.undefined

      it 'does not remove other entries', ->
        expect(@entry_store.get user: @user_2).to.eql @user_2


    context 'ids query given', ->

      beforeEach ->
        @entry_store.delete username: 'one'

      it 'removes the entry described by the given query from this store', ->
        expect(@entry_store.get user: @user_1).to.be.undefined

      it 'does not remove other entries', ->
        expect(@entry_store.get user: @user_2).to.eql @user_2



  describe 'delete_many', ->

    beforeEach ->
      @entry_store.add_many [@user_1, @user_2, @user_3]


    context 'entries query given', ->

      beforeEach ->
        @entry_store.delete_many users: [@user_1, @user_2]

      it 'removes the entries described by the given query from this store', ->
        expect(@entry_store.get user: @user_1).to.be.undefined
        expect(@entry_store.get user: @user_2).to.be.undefined

      it 'does not remove other entries', ->
        expect(@entry_store.get user: @user_3).to.eql @user_3


    context 'ids query given', ->

      beforeEach ->
        @entry_store.delete_many usernames: ['one', 'two']

      it 'removes the entries described by the given query from this store', ->
        expect(@entry_store.get user: @user_1).to.be.undefined
        expect(@entry_store.get user: @user_2).to.be.undefined

      it 'does not remove other entries', ->
        expect(@entry_store.get user: @user_3).to.eql @user_3



  describe 'entry_changes', ->

    beforeEach ->
      @entry_store.add @user_1
      @user = @entry_store.get user: @user_1


    context 'the entry is changed', ->

      beforeEach ->
        @user.text = 'new text 1'
        @result = @entry_store.entry_changes @user

      it 'returns the changed entries, only their changed fields', ->
        expect(@result).to.eql {username: 'one', text: 'new text 1'}


    context 'the entry is unchanged', ->

      beforeEach ->
        @result = @entry_store.entry_changes @user

      it 'returns false', ->
        expect(@result).to.be.false



  describe 'entries_changes', ->

    beforeEach ->
      @entry_store.add_many [@user_1, @user_2, @user_3]
      @user_1_changed = @entry_store.get user: @user_1
      @user_2_changed = @entry_store.get user: @user_2
      @user_3_changed = @entry_store.get user: @user_3


    context 'some entries are changed', ->

      beforeEach ->
        @user_1_changed.text = 'new text 1'
        @user_3_changed.text = 'new text 3'
        @result = @entry_store.entries_changes [@user_1_changed,
                                                @user_3_changed]

      it 'returns the changed fields of the entries', ->
        expect(@result).to.eql [ { username: 'one', text: 'new text 1' }
                                 { username: 'three', text: 'new text 3' } ]

    context 'all entries are unchanged', ->

      beforeEach ->
        @result = @entry_store.entries_changes [@user_1_changed,
                                                @user_2_changed]

      it 'returns an empty array', ->
        expect(@result).to.eql []


  describe 'get', ->

    describe 'normal behavior', ->

      beforeEach ->
        @entry_store.add @user_1

      it 'returns a modifiable version of existing unmodified entries', ->
        expect(@entry_store.get user: @user_1).to.be.an.editable_version_of @user_1

      it 'returns modified entries in the modified version', ->
        user = @entry_store.get user: @user_1
        user.text = 'new text'
        result = @entry_store.get user: @user_1
        expect(result).to.have.property 'username', 'one'
        expect(result).to.have.property 'text', 'new text'

      it 'returns undefined if the entry is not cached', ->
        expect(@entry_store.get id: 99).to.be.undefined


    describe 'partial entries enabled', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @entry_store = new slender_data.EntryStore @test_params

      context 'partial entry in cache', ->

        beforeEach ->
          @entry_store.add @partial_user_1

        it 'returns the partial entry when partial entries are allowed', ->
          expect(@entry_store.get username: 'one').to.eql @partial_user_1

        it 'returns nothing if full entries are requested', ->
          expect(@entry_store.get username: 'one', full: yes).to.be.undefined

      context 'full entry in cache', ->

        beforeEach ->
          @entry_store.add @full_user_1

        it 'returns the full entry when requesting a partial entry', ->
          expect(@entry_store.get username: 'one').to.eql @full_user_1

        it 'returns the full entry when requesting a full entry', ->
          expect(@entry_store.get username: 'one', full: yes).to.eql @full_user_1


  describe 'get_many', ->

    describe 'normal functionality', ->

      beforeEach ->
        @entry_store.add_many [@user_1, @user_2]
        @result = @entry_store.get_many users: [@user_1, @user_2, @user_3]

      it 'returns the cached entries', ->
        expect(@result.found).to.eql [@user_1, @user_2]

      it 'returns editable versions of the cached entries', ->
        expect(@result.found[0]).to.be.an.editable_version_of @user_1

      it 'returns the missing entries as a MultiQuery', ->
        expect(@result.missing).to.be.an.instanceof slender_data.MultiQuery
        expect(@result.missing.selector_value()).to.eql [@user_3]

      it 'returns modified snippets in the modified form', ->
        user_1_modified = @entry_store.get user: @user_1
        user_1_modified.text = 'new text'
        result = @entry_store.get_many users: [@user_1]
        expect(result.found).to.eql [user_1_modified]
        expect(result.found).to.not.eql [@user_1]


    describe 'partial entry functionality', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @entry_store = new slender_data.EntryStore @test_params
        @entry_store.add_many [@partial_user_1, @full_user_2]

      context 'full entries requested', (done) ->

        beforeEach ->
          @result = @entry_store.get_many
            usernames: ['one', 'two', 'three'],
            full: yes

        it 'returns only the full entries', ->
          expect(@result.found).to.eql [@full_user_2]

        it 'does not return partial entries', ->
          expect(@result.missing.to_hash().usernames).to.include 'one'

        it 'does not return missing entries', ->
          expect(@result.missing.to_hash().usernames).to.include 'three'


      context 'partial entries allowed', ->

        beforeEach ->
          @result = @entry_store.get_many usernames: ['one', 'two', 'zonk'], full: no

        it 'returns all requested entries from the cache', ->
          expect(@result.found).to.eql [@partial_user_1, @full_user_2]

        it 'reports missing entries', ->
          expect(@result.missing.to_hash().usernames).to.eql ['zonk']


    describe 'versioning', ->

      beforeEach ->
        @test_params.versioning = yes
        @entry_store = new slender_data.EntryStore @test_params
        @entry_store.add_many [@user_1_outdated, @user_2]
        @result = @entry_store.get_many users: [@user_1, @user_2]

      it 'returns all up-to-date requested entries', ->
        expect(@result.found).to.eql [@user_2]

      it 'reports outdated entries as missing', ->
        expect(@result.missing).to.eql_query users: [@user_1]

