#= require spec_helper


# Domain-specifc matcher for checking if the PersistenceManager
# has the given user in the cache (or not).
chai.Assertion.addMethod 'cached_user', (user) ->
  cached_user = @_obj.get user: user
  if @__flags.negate
    # No cached user expected.
    new chai.Assertion(cached_user,
                       'Should not have cached user').to.be.undefined
  else
    # The given cached user expected.
    new chai.Assertion(cached_user,
                       'Should have cached user').to.be.eql user
    new chai.Assertion(cached_user,
                       'Should have cached user').to.not.be.equal user


# Domain-specifc matcher for checking if the tested object is an array
# of elements with at least the given properties.
chai.Assertion.addMethod 'entries_with_matching_properties', (attr_name, expected_values) ->
  expect(@_obj).to.have.length expected_values.length
  obj_values = (obj[attr_name] for obj in @_obj)
  for expected_value in expected_values
    expect(obj_values).to.include expected_value


describe 'PersistenceManager', ->

  beforeEach ->
    define_test_data this
    @persistence_manager = new slender_data.PersistenceManager @test_params


  describe 'add_to_cache', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @persistence_manager.add_to_cache @user_1

      it 'adds the described entry to the serverside cache', ->
        expect(@persistence_manager).to.have.cached_user @user_1

      it 'returns nothing', ->
        expect(@result).to.be.undefined

    describe 'return_editable_version option', ->

      it 'returns an editable version of the given entry when set', ->
        result = @persistence_manager.add_to_cache @user_1, yes
        expect(result).to.be.an.editable_version_of @user_1

      it 'returns nothing if not set', ->
        result = @persistence_manager.add_to_cache @user_1, no
        expect(result).to.be.undefined


  describe 'add_many_to_cache', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @persistence_manager.add_many_to_cache [@user_1, @user_2]

      it 'adds all described entries to the chache', ->
        expect(@persistence_manager).to.have.cached_user @user_1
        expect(@persistence_manager).to.have.cached_user @user_2

      it 'returns nothing', ->
        expect(@result).to.be.undefined


    describe 'return_editable_version option', ->

      it 'returns editable versions of the given entries when set', ->
        result = @persistence_manager.add_many_to_cache [@user_1, @user_2], yes
        expect(result[0]).to.be.an.editable_version_of @user_1
        expect(result[1]).to.be.an.editable_version_of @user_2

      it 'returns nothing if not set', ->
        result = @persistence_manager.add_many_to_cache [@user_1, @user_2], no
        expect(result).to.be.undefined



  describe 'create', ->

    beforeEach ->
      @server_create_api = sinon.stub(@persistence_manager.server_api, 'create')
                                .yields user: @user_1
      @result = @persistence_manager.create @user_1

    it 'creates the entry on the server', ->
      expect(@server_create_api).to.have.been.calledOnceWith @user_1

    it 'adds the server response to the cache', ->
      expect(@persistence_manager).to.have.cached_user @user_1

    it 'returns nothing', ->
      expect(@result).to.be.undefined

    it 'calls the given callback with an editable version of the created entry when done', (done) ->
      @persistence_manager.create @user_1, (result) =>
        expect(result).to.be.an.editable_version_of @user_1
        done()


  describe 'create_many', ->

    beforeEach ->
      @server_create_many_api = sinon.stub(@persistence_manager.server_api, 'create_many')
                                     .yields users: [@user_1, @user_2]
      @result = @persistence_manager.create_many [@user_1, @user_2]

    it 'creates the entries on the server', ->
      expect(@server_create_many_api).to.have.been.calledOnceWith [@user_1, @user_2]

    it 'adds the server response to the cache', ->
      expect(@persistence_manager).to.have.cached_user @user_1
      expect(@persistence_manager).to.have.cached_user @user_2

    it 'returns nothing', ->
      expect(@result).to.be.undefined

    it 'calls the given callback when done', (done) ->
      @persistence_manager.create_many [@user_1], done



  describe 'delete', ->

    beforeEach (done) ->
      @persistence_manager.add_to_cache @user_1
      @server_delete_api = sinon.stub(@persistence_manager.server_api, 'delete')
                                .yields username: 'one'
      @persistence_manager.delete @user_1, (@result) =>
        done()

    it 'removes the given entry from the cache', ->
      expect(@persistence_manager).to.not.have.cached_user @user_1

    it 'removes the given entry from the server', ->
      expect(@server_delete_api).to.have.been.calledOnceWith @user_1

    it 'calls the given callback with a description of the deleted entry', ->
      expect(@result).to.eql username: 'one'


  describe 'delete_many', ->

    beforeEach (done) ->
      @persistence_manager.add_many_to_cache [@user_1, @user_2]
      @server_delete_many_api = sinon.stub(@persistence_manager.server_api, 'delete_many')
                                     .yields usernames: ['one', 'two']
      @persistence_manager.delete_many [@user_1, @user_2], (@result) => done()

    it 'removes the given entries from the cache', ->
      expect(@persistence_manager).to.not.have.cached_user @user_1
      expect(@persistence_manager).to.not.have.cached_user @user_2

    it 'removes the given entry from the server', ->
      expect(@server_delete_many_api).to.have.been.calledOnceWith [@user_1, @user_2]

    it 'calls the given callback with a description of the deleted entries', ->
      expect(@result).to.eql usernames: ['one', 'two']



  describe 'get', ->

    describe 'normal functionality', ->

      beforeEach ->
        @persistence_manager.add_to_cache @user_1

      it 'returns the described entry from the cache if it exists there', ->
        expect(@persistence_manager).to.have.cached_user @user_1

      it 'returns undefined if the described entry is not cached', ->
        expect(@persistence_manager.get user: @user_2).to.be.undefined


    describe 'partial/full functionality', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @persistence_manager = new slender_data.PersistenceManager @test_params

      context 'partial entry in cache', ->

        beforeEach ->
          @persistence_manager.add_to_cache @partial_user_1

        it 'returns the cached partial entry when partial entries are allowed', ->
          result = @persistence_manager.get username: 'one', full: no
          expect(result).to.eql @partial_user_1

        it 'returns undefined when full entries are requested', ->
          result = @persistence_manager.get username: 'one', full: yes
          expect(result).to.be.undefined


      context 'full entry in the cache', ->

        beforeEach ->
          @persistence_manager.add_to_cache @full_user_1

        context 'partial entry requested', ->

          beforeEach ->
            @result = @persistence_manager.get username: 'one', full: no

          it 'returns the cached full entry', ->
            expect(@result).to.eql @full_user_1

        context 'full entry requested', ->

          beforeEach ->
            @result = @persistence_manager.get username: 'one', full: yes

          it 'returns the cached full entry', ->
            expect(@result).to.eql @full_user_1



  describe 'get_many', ->

    describe 'normal functionality', ->

      beforeEach ->
        @persistence_manager.add_many_to_cache [@user_1, @user_3]
        @result = @persistence_manager.get_many users: [@user_1, @user_2, @user_3]

      it 'returns the cached entries', ->
        expect(@result.found).to.eql [@user_1, @user_3]

      it 'returns the ids of the missing entries', ->
        expect(@result.missing).to.be.an.instanceof slender_data.MultiQuery
        expect(@result.missing.ids()).to.eql ['two']


    describe 'partial/full functionality', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @persistence_manager = new slender_data.PersistenceManager @test_params
        @persistence_manager.add_many_to_cache [@partial_user_1, @full_user_2]

      context 'partial entries requested', ->

        beforeEach ->
          @result = @persistence_manager.get_many
            usernames: ['one', 'two', 'three'], full: no

        it 'returns all matching cached entries as found', ->
          expect(@result.found).to.eql [@partial_user_1, @full_user_2]

        it 'returns the ids of missing entries as missing', ->
          expect(@result.missing).to.eql_query usernames: ['three'], full: no

      context 'full entries requested', ->

        beforeEach ->
          @result = @persistence_manager.get_many
            usernames: ['one', 'two', 'three'], full: yes

        it 'returns only full entries as found', ->
          expect(@result.found).to.eql [@full_user_2]

        it 'returns partial and missing entries as missing', ->
          expect(@result.missing).to.eql_query usernames: ['one', 'three'], full: yes


    describe 'versioning', ->

       beforeEach ->
         @test_params.versioning = yes
         @persistence_manager = new slender_data.PersistenceManager @test_params
         @persistence_manager.add_many_to_cache [@user_1_outdated, @user_2]
         @result = @persistence_manager.get_many users: [@user_1, @user_2]

        it 'returns only results of the right version', ->
          expect(@result.found).to.eql [@user_2]
          expect(@result.missing).to.eql_query users: [@user_1]


  describe 'load', ->

    beforeEach ->
      @server_read_api = sinon.stub @persistence_manager.server_api, 'read'


    describe 'normal behavior', ->

      context 'described entry is in the cache', ->

        beforeEach (done) ->
          @persistence_manager.add_to_cache @user_1
          @persistence_manager.load user: @user_1, (@result) =>
            done()

        it 'returns the entry directly from the cache', ->
          expect(@result).to.eql @user_1

        it 'does not query the server', ->
          expect(@persistence_manager.server_api.read).to.not.have.been.called

      context 'described entry is currently loading', ->

        beforeEach ->
          @persistence_manager.loading_tracker.start_loading @user_1
          @callback = sinon.stub()
          @persistence_manager.load user: @user_1, @callback

        it 'does not make another load operation to the server', ->
          expect(@server_read_api).to.not.have.been.called

        it 'calls the callback with the entry when the entry returns from the server', ->
          expect(@callback).to.not.have.been.called
          @persistence_manager.loading_tracker.entry_arrived @user_1
          expect(@callback).to.have.been.calledOnceWith @user_1

        it 'adds the loaded entry to the clientside cache', ->
          @persistence_manager.loading_tracker.entry_arrived @user_1
          expect(@persistence_manager).to.have.cached_user @user_1


      context 'described entry is not in the cache', ->

        beforeEach (done) ->
          @server_read_api.yields(user: @user_1)
          @persistence_manager.load user: @user_1, (@result) =>
            done()

        it 'loads the entry from the server', ->
          server_read_api = @persistence_manager.server_api.read
          expect(server_read_api).to.have.been.calledOnceWithQuery user: @user_1

        it 'adds the loaded entry to the cache', ->
          expect(@persistence_manager).to.have.cached_user @user_1

        it 'returns an editable copy of the entry loaded from the server', ->
          expect(@result).to.be.an.editable_version_of @user_1


    describe 'partial entry functionality', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @persistence_manager = new slender_data.PersistenceManager @test_params
        @server_read_api = sinon.stub(@persistence_manager.server_api, 'read')
                                .yields(user: @full_user_1)

      context 'cache contains partial entry', ->

        beforeEach ->
          @persistence_manager.add_to_cache @partial_user_1

        context 'full entry requested', (done) ->

          beforeEach (done) ->
            @persistence_manager.load {username: 'one', full: yes}, (@result) =>
              done()

          it 'loads the requested entry from the server', ->
              expect(@server_read_api).to.have.been.calledOnce
              expect(@server_read_api.args[0][0].to_hash()).to.have.property 'username',
                                                                             'one'

          it 'replaces the cached partial entry with the loaded full entry', ->
            expect(@persistence_manager.get username: 'one').to.eql @full_user_1

        context 'partial entry requested', ->

          beforeEach (done) ->
            @persistence_manager.load {username: 'one', full: no}, (@result) =>
              done()

          it 'returns the entry from the cache', ->
            expect(@result).to.eql @partial_user_1

          it 'does not load data from the server', ->
            expect(@server_read_api).to.not.have.been.called

      context 'cache contains full entry', ->

        beforeEach ->
          @persistence_manager.add_to_cache @full_user_1

        context 'full entry requested', ->

          beforeEach (done) ->
            @persistence_manager.load {username: 'one', full: yes}, (@result) =>
              done()

          it 'returns the entry from the cache', ->
            expect(@result).to.eql @full_user_1

          it 'does not load data from the server', ->
            expect(@server_read_api).to.not.have.been.called

        context 'partial entry requested', ->

          beforeEach (done) ->
            @persistence_manager.load {username: 'one', full: no}, (@result) =>
              done()

          it 'returns the full entry from the cache', ->
            expect(@result).to.eql @full_user_1

          it 'does not load data from the server', ->
            expect(@server_read_api).to.not.have.been.called


    describe 'versioning', ->

      beforeEach ->
        @test_params.versioning = yes
        @persistence_manager = new slender_data.PersistenceManager @test_params
        @server_read_api = sinon.stub(@persistence_manager.server_api, 'read')

      context 'cache contains current entry', ->

        beforeEach (done) ->
          @persistence_manager.add_to_cache @user_1
          @persistence_manager.load user: @user_1, (@result) => done()

        it 'returns the cached entry', ->
          expect(@result).to.eql @user_1

        it 'does not load anything from the server', ->
          expect(@server_read_api).to.not.have.been.called

      context 'cache contains outdated entry', ->

        beforeEach (done) ->
          @persistence_manager.add_to_cache @user_1_outdated
          @server_read_api.yields user: @user_1
          @persistence_manager.load user: @user_1, (@result) => done()

        it 'does not return the cached entry', ->
          expect(@result).to.not.eql @user_1_outdated

        it 'loads the current entry from the server', ->
          expect(@server_read_api).to.have.been.calledOnce

        it 'returns the current entry loaded from the server', ->
          expect(@result).to.eql @user_1

        it 'updates the cache with the newly loaded entry version', ->
          expect(@persistence_manager).to.have.cached_user @user_1



  describe 'load_many', ->

    beforeEach ->
      @server_read_many_api = sinon.stub @persistence_manager.server_api, 'read_many'


    describe 'normal functionality', ->

      context 'all snippets are in the cache', ->

        beforeEach ->
          @persistence_manager.add_many_to_cache [@user_1, @user_2]

        it 'returns all requested snippets', (done) ->
          @persistence_manager.load_many users: [@user_1, @user_2], (users) =>
            expect(users).to.eql [@user_1, @user_2]
            done()

        it 'does not query the server', (done) ->
          @persistence_manager.load_many users: [@user_1, @user_2], =>
            expect(@server_read_many_api).to.not.have.been.called
            done()

        it 'works without a given callback', ->
          @persistence_manager.load_many users: [@user_1]


      context 'some snippets are not cached', ->

        beforeEach ->
          @persistence_manager.add_many_to_cache [@user_1, @user_3]
          @user_2_from_server = _.clone @user_2
          @server_read_many_api.yields users: [@user_2_from_server]

        it 'returns all requested snippets', (done) ->
          @persistence_manager.load_many users: [@user_1,
                                                 @user_2,
                                                 @user_3], (users) =>
            expect(users).to.have.entries_with_matching_properties 'username',
                                                                   [ 'one',
                                                                     'two',
                                                                     'three' ]
             done()

        it 'loads the missing snippets from the server', (done) ->
          @persistence_manager.load_many users: [@user_1, @user_2, @user_3], (users) =>
            expect(@server_read_many_api).to.have.been.calledOnceWithQuery users: [@user_2]
            done()

        it 'adds the loaded entries to the cache', ->
          @persistence_manager.load_many users: [@user_2], (users) =>
            expect(users[0]).to.eql @user_2
            expect(users[0]).to.not.equal @user_2_from_server


    describe 'partial entry functionality', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @persistence_manager = new slender_data.PersistenceManager @test_params
        @persistence_manager.add_many_to_cache [@partial_user_1, @full_user_2]
        @server_read_many_api = sinon.stub(@persistence_manager.server_api, 'read_many')
                                     .yields(users: [ @full_user_1, @full_user_3 ])

      context 'full entries requested', (done) ->

        beforeEach (done) ->
          @persistence_manager.load_many {usernames: ['one', 'two', 'three'], full: yes},
                                         (@result) => done()

        it 'loads the partial entries from the server', ->
          expect(@server_read_many_api).to.have.been.calledOnce
          server_args = @server_read_many_api.args[0][0]
          expect(server_args.to_hash().usernames).to.eql ['one', 'three']

        it 'replaces the cached partial entries with the loaded full entries', ->
          expect(@persistence_manager.get username: 'one').to.eql @full_user_1
          expect(@persistence_manager.get username: 'two').to.eql @full_user_2

        it 'returns the full entries', ->
          expect(@result).to.have.entries_with_matching_properties 'email',
                                                                   [ 'one@example.com',
                                                                     'two@example.com',
                                                                     'three@example.com' ]

      context 'partial entries requested', ->

        beforeEach (done) ->
          @persistence_manager.load_many {usernames: ['one', 'two'], full: no},
                                         (@result) => done()

        it 'returns the entries from the cache', ->
          expect(@result).to.eql [@partial_user_1, @full_user_2]

        it 'does not load data from the server', ->
          expect(@server_read_many_api).to.not.have.been.called


    describe 'versioning', ->

      beforeEach (done) ->
        @test_params.versioning = yes
        @persistence_manager = new slender_data.PersistenceManager @test_params
        @persistence_manager.add_many_to_cache [@user_1_outdated, @user_2]
        @server_read_many_api = sinon.stub(@persistence_manager.server_api,
                                           'read_many')
                                     .yields(users: [@user_1])
        @persistence_manager.load_many users: [@user_1, @user_2],
                                       (@result) => done()

      it 'returns all requested entries', ->
        expect(@result).to.eql [@user_2, @user_1]

      it 'loads only outdated entries from the server', ->
        expect(@server_read_many_api).to.have.been.calledOnceWithQuery users: [@user_1]

      it 'adds the loaded entries to the cache', ->
        expect(@persistence_manager.get username: 'one').to.eql @user_1


    describe 'loading tracking', ->

      context 'some entries are currently loading', ->

        beforeEach ->
          @persistence_manager.loading_tracker.start_loadings users: [@user_1, @user_2]
          @callback = sinon.stub()
          @persistence_manager.load_many users: [@user_1, @user_2], @callback

        it 'does not make another load operation to the server', ->
          expect(@server_read_many_api).to.not.have.been.called

        it 'calls the callback with the entries when they arrive', ->
          expect(@callback).to.not.have.been.called
          @persistence_manager.loading_tracker.entries_arrived [@user_1, @user_2]
          expect(@callback).to.have.been.calledOnceWith [@user_1, @user_2]

        it 'adds the loaded result to the cache', ->
          @persistence_manager.loading_tracker.entries_arrived [@user_1, @user_2]
          expect(@persistence_manager).to.have.cached_user @user_1
          expect(@persistence_manager).to.have.cached_user @user_2

      context 'nothing is loading right now', ->

        beforeEach ->
          @persistence_manager.load_many users: [@user_1], @callback
          @persistence_manager.load_many users: [@user_1], @callback

        it 'marks entries that are now being loaded as loading', ->
          expect(@persistence_manager.loading_tracker).to.have.loading @user_1

        it 'calls the server API only once for repeated requests to load data', ->
          expect(@server_read_many_api).to.have.been.calledOnce


      context 'the server response arrives', ->

        beforeEach ->
          @server_read_many_api.yields users: [@user_1]
          @persistence_manager.load_many users: [@user_1]

        it 'tells the loading tracker that the entries have arrived', ->
          expect(@persistence_manager.loading_tracker).to.not.have.loading @user_1


      describe 'all options working together', ->

        beforeEach ->
          @persistence_manager.add_to_cache @user_1
          @persistence_manager.loading_tracker.start_loadings users: [@user_2]
          @callback = sinon.stub()
          @server_read_many_api.yields users: [@user_3]
          @persistence_manager.load_many users: [@user_1, @user_2, @user_3],
                                         @callback
          @persistence_manager.loading_tracker.entries_arrived [@user_2]

        it 'returns all requested entries', ->
          expect(@callback).to.have.been.calledOnce
          callback_args = @callback.args[0][0]
          expect(callback_args).to.be.entries_with_matching_properties 'username',
                                                                       [ 'one',
                                                                         'two',
                                                                         'three' ]

        it 'loads only the not currently loading entries from the server', ->
          expect(@server_read_many_api).to.have.been.calledOnceWithQuery users: [@user_3]



  describe 'merge', ->

    beforeEach ->
      @merged_user_data = email: 'merged@example.com'
      @merged_user_server_response = username: 'merged server', timestamp: 1
      @persistence_manager.add_many_to_cache [@user_1, @user_2, @user_3]
      @callback = sinon.stub()
      @server_merge_api = sinon.stub(@persistence_manager.server_api, 'merge')
                               .yields(user: @merged_user_server_response)
      @persistence_manager.merge [@user_1.username, @user_2.username],
                                 @merged_user_data,
                                 @callback

    it 'performes the merge on the server', ->
      expect(@server_merge_api).to.have.been.calledOnce

    it 'removes the old users from the client cache', ->
      expect(@persistence_manager).to.not.have.cached_user @user_1
      expect(@persistence_manager).to.not.have.cached_user @user_2
      expect(@persistence_manager).to.have.cached_user @user_3

    it 'adds the user provided by the server to the client cache', ->
      cached_user = @persistence_manager.get user: @merged_user_server_response
      expect(cached_user).to.eql @merged_user_server_response

    it 'calls the given callback when done', ->
      expect(@callback).to.have.been.calledOnce

    it 'returns the editable user in the callback', ->
      callback_args = @callback.args[0][0]
      expect(callback_args).to.be.an.editable_version_of @merged_user_server_response



  describe 'update', ->

    beforeEach ->
      @persistence_manager.add_to_cache @user_1
      @user = @persistence_manager.get user: @user_1
      @server_update_api = sinon.stub(@persistence_manager.server_api, 'update')
                                                          .yields user: @user


    context 'nothing is changed', ->

      beforeEach (done) ->
        @persistence_manager.update @user, (@result) => done()

      it 'does not make a call to the server', ->
        expect(@server_update_api).to.not.have.been.called

      it 'calls the callback with the given entry', ->
        expect(@result).to.eql @user_1

      it 'works without a callback provided', ->
        @persistence_manager.update @user


    context 'entry is changed', ->

      beforeEach (done) ->
        @user.email = 'new_one@example.com'
        @persistence_manager.update @user, (@result) => done()

      it 'calls the server with only the changed attributes as payload', ->
        expect(@server_update_api).to.have.been.calledOnceWith
          username: 'one', email: 'new_one@example.com'

      it 'calls the callback with the editable version of the updated entry', ->
        expect(@result).to.be.an.editable_version_of @user

      it 'stores the now persisted changes on the client', (done) ->
        @persistence_manager.update @user, =>
          @persistence_manager.update @user, =>
            expect(@persistence_manager.server_api.update).to.have.been.calledOnce
            done()

      it 'works without a given callback', ->
        @persistence_manager.update @user



  describe 'update_many', ->

    beforeEach ->
      @persistence_manager.add_many_to_cache [@user_1, @user_2, @user_3]
      @user_1_changed = @persistence_manager.get user: @user_1
      @user_2_changed = @persistence_manager.get user: @user_2
      @user_3_changed = @persistence_manager.get user: @user_3
      @server_update_many_api = sinon.stub @persistence_manager.server_api,
                                           'update_many'

    context 'nothing is changed', ->

      beforeEach (done) ->
        @server_update_many_api.yields users: [ @user_1_changed,
                                                @user_2_changed ]
        @persistence_manager.update_many [@user_1_changed, @user_2_changed],
                                         (@result) => done()

      it 'does not call the server', ->
        expect(@server_update_many_api).to.not.have.been.called

      it 'calls the callback with the server response', ->
        expect(@result).to.eql users: [@user_1_changed, @user_2_changed]

      it 'works without a given callback', ->
        @persistence_manager.update_many [@user_1_changed]


    context 'some entries are changed', ->

      beforeEach (done) ->
        @user_1_changed.email = 'new_one@example.com'
        @user_3_changed.timestamp = 3
        @server_update_many_api.yields users: [ @user_1_changed,
                                                @user_3_changed  ]
        @persistence_manager.update_many [ @user_1_changed
                                           @user_2_changed
                                           @user_3_changed ],
                                         (@result) => done()

      it 'calls the server with only the changed attributes as payload', ->
          expect(@server_update_many_api).to.have.been.calledOnce
          update_args = @server_update_many_api.args[0][0]
          expect(update_args[0]).to.eql username: 'one', email: 'new_one@example.com'
          expect(update_args[1]).to.eql username: 'three', timestamp: 3

      it 'calls the callback with the server response', ->
        expect(@result).to.eql users: [@user_1_changed, @user_3_changed]

      it 'stores the now persisted changes on the client', (done) ->
        @persistence_manager.update_many [ @user_1_changed
                                           @user_2_changed
                                           @user_3_changed ], =>
          @persistence_manager.update_many [ @user_1_changed
                                             @user_2_changed
                                             @user_3_changed ], =>
            expect(@server_update_many_api).to.have.been.calledOnce
            done()

      it 'works without a given callback', ->
        @persistence_manager.update_many [@user_1]

