#= require spec_helper
#= require slender_data/queries/multi_query


describe 'slender_data.MultiQuery', ->

  beforeEach ->
    define_test_data this


  describe 'constructor', ->

    it 'allows to provide ids using their generic name "ids"', ->
      multi_query = new slender_data.MultiQuery ids: [1, 2], @test_params
      expect(multi_query.ids()).to.eql [1, 2]

    it 'allows to provide ids using their domain specific name (i.e. "usernames")', ->
      multi_query = new slender_data.MultiQuery usernames: [1, 2], @test_params
      expect(multi_query.ids()).to.eql [1, 2]

    it 'allows to provide entries using their generic name "entries"', ->
      multi_query = new slender_data.MultiQuery entries: [ {username: 'one'},
                                                           {username: 'two'} ],
                                                @test_params
      expect(multi_query.entries()).to.eql [ {username: 'one'},
                                             {username: 'two'} ]

    it 'allows to provide entries using their domain specific name (i.e. "users")', ->
      multi_query = new slender_data.MultiQuery users: [ {username: 'one'},
                                                         {username: 'two'} ],
                                                @test_params
      expect(multi_query.entries()).to.eql [ {username: 'one'},
                                             {username: 'two'} ]

    it 'throws an exception for unknown query selectors', ->
      expect(-> new slender_data.MultiQuery [1,2]).to.throw()
      expect(-> new slender_data.MultiQuery zonk: [1,2]).to.throw()



  describe 'empty', ->

    it 'returns true if the multi-query describes no entries', ->
      multi_query = new slender_data.MultiQuery ids: [], @test_params
      expect(multi_query.empty()).to.be.true

    it 'returns false if the multi-query describes at least one snippet', ->
      multi_query = new slender_data.MultiQuery ids: [1], @test_params
      expect(multi_query.empty()).to.be.false



  describe 'entries', ->

    it 'throws an exception for an ids query', ->
      query = new slender_data.MultiQuery usernames: ['one'], @test_params
      expect(-> query.entries()).to.throw()

    it 'returns the entries for the given entries query', ->
      users = [{username: 'one'}]
      query = new slender_data.MultiQuery users: users, @test_params
      result = query.users()
      expect(result).to.eql users



  describe 'ids', ->

    it 'returns the ids of the given ids query', ->
      query = new slender_data.MultiQuery usernames: ['one', 'two'], @test_params
      expect(query.usernames()).to.eql ['one', 'two']

    it 'returns the id of the given entry query', ->
      query = new slender_data.MultiQuery users: [ {username: 'one'}
                                                   {username: 'two'} ],
                                          @test_params
      expect(query.ids()).to.eql ['one', 'two']



  describe 'replicate', ->

    beforeEach ->
      @test_params.versioning = yes
      @test_params.partial_entries = yes
      multi_query = new slender_data.MultiQuery users: [@user_1], @test_params
      multi_query.full = yes
      @result = multi_query.replicate [@user_2]

    it 'creates a new MultiQuery', ->
      expect(@result).to.be.an.instanceof slender_data.MultiQuery

    it 'makes the new MultiQuery have the same selector name', ->
      expect(@result.selector_name()).to.eql 'users'

    it 'makes the new MultiQuery have the given selector values', ->
      expect(@result.selector_value()).to.eql [@user_2]

    it 'copies the "partial_entries" configuration setting to the new MultiQuery', ->
      expect(@result.partial_entries).to.be.true

    it 'copies the "versioning" configuration setting to the new MultiQuery', ->
      expect(@result.versioning).to.be.true

    it 'copies the "full" query parameter to the new MultiQuery', ->
      expect(@result.full).to.be.true



  describe 'single_selector_name', ->

    it 'returns the custom entry name if the custom entries name is given', ->
      @test_params.entry_name = 'entry'
      @test_params.entries_name = 'entries'
      multi_query = new slender_data.MultiQuery entries: [{id: 1}], @test_params
      expect(multi_query.single_selector_name()).to.equal 'entry'

    it 'it truncates the last character for any other selector name', ->
      multi_query = new slender_data.MultiQuery ids: [1], @test_params
      expect(multi_query.single_selector_name()).to.equal 'id'



  describe 'transpose', ->

    it 'converts the given ids MultiQuery into an array of id Queries', ->
      multi_query = new slender_data.MultiQuery usernames: ['one', 'two', 'three'],
                                                @test_params
      result = multi_query.transpose()
      expect(result[0]).to.eql_query username: 'one'
      expect(result[1]).to.eql_query username: 'two'
      expect(result[2]).to.eql_query username: 'three'

    it 'converts the given entries MultiQuery into an array of entry Queries', ->
      multi_query = new slender_data.MultiQuery
        users: [{username: 'one'}, {username: 'two'}, {username: 'three'}],
        @test_params
      result = multi_query.transpose()
      expect(result.length).to.eql 3
      expect(result[0]).to.eql_query user: {username: 'one'}
      expect(result[1]).to.eql_query user: {username: 'two'}
      expect(result[2]).to.eql_query user: {username: 'three'}

