#= require spec_helper


describe 'QueryCache', ->

  beforeEach ->
    define_test_data this
    @query_cache = new slender_data.QueryCache @test_params


  describe 'add', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @query_cache.add @user_1

      it 'adds the given entry', ->
        expect(@query_cache.cache.length()).to.equal 1
        expect(@query_cache.get username: 'one').to.equal @user_1

      it 'returns nothing', ->
        expect(@result).to.be.undefined


    describe 'partial/full behavior', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @query_cache = new slender_data.QueryCache @test_params

      it 'replaces an existing partial entry with the given full entry', ->
        @query_cache.add @partial_user_1
        @query_cache.add @full_user_1
        expect(@query_cache.get username: 'one').to.eql @full_user_1

      it 'does not replace an existing full entry with the given partial entry', ->
        @query_cache.add @full_user_1
        @query_cache.add @partial_user_1
        expect(@query_cache.get username: 'one').to.eql @full_user_1



  describe 'add_clone', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @query_cache.add_clone @user_1

      it 'adds a copy of the given entry', ->
        expect(@query_cache.length()).to.equal 1
        expect(@query_cache.get user: @user_1).to.be.an.editable_version_of @user_1

      it 'returns a copy of the added entry', ->
        expect(@result).to.be.an.editable_version_of @user_1


    describe 'partial/full behavior', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @query_cache = new slender_data.QueryCache @test_params

      it 'replaces an existing partial entry with the given full entry', ->
        @query_cache.add @partial_user_1
        @query_cache.add_clone @full_user_1
        expect(@query_cache.get username: 'one').to.eql @full_user_1

      context 'full entry cached, partial entry given', ->

        beforeEach ->
          @query_cache.add @full_user_1
          @result = @query_cache.add_clone @partial_user_1, true

        it 'does not replace the existing entry', ->
          expect(@query_cache.get username: 'one').to.eql @full_user_1

        it 'returns the existing full entry anyways', ->
          expect(@result).to.eql @full_user_1



    describe 'add_clones', ->

      beforeEach ->
        @result = @query_cache.add_clones [@user_1, @user_2]

      it 'adds copies of the given entries to the chache', ->
        expect(@query_cache.length()).to.equal 2
        cache_content = @query_cache.get_many users: [@user_1, @user_2]
        expect(cache_content.found[0]).to.be.an.editable_version_of @user_1
        expect(cache_content.found[1]).to.be.an.editable_version_of @user_2

      it 'returns copies of the added entries', ->
        expect(@result).to.eql [@user_1, @user_2]



  describe 'add_many', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @query_cache.add_many [@user_1, @user_2]

      it 'adds the given entries', ->
        expect(@query_cache.length()).to.equal 2
        expect(@query_cache.get user: @user_1).to.equal @user_1
        expect(@query_cache.get user: @user_2).to.equal @user_2

      it 'returns undefined', ->
        expect(@result).to.be.undefined


    describe 'partial/full behavior', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @query_cache = new slender_data.QueryCache @test_params
        @query_cache.add_many [@partial_user_1, @full_user_2]
        @query_cache.add_many [@full_user_1, @partial_user_2]

      it 'replaces an existing partial entry with the given full entry', ->
        expect(@query_cache.get username: 'one').to.eql @full_user_1

      it 'does not replace an existing full entry with the given partial entry', ->
        expect(@query_cache.get username: 'two').to.eql @full_user_2



  describe 'add_many_as_clones', ->

    describe 'normal behavior', ->

      beforeEach ->
        @result = @query_cache.add_many_as_clones [@user_1, @user_2]

      it 'adds clones of the given entries', ->
        expect(@query_cache.length()).to.equal 2
        expect(@query_cache.get user: @user_1).to.be.an.editable_version_of @user_1
        expect(@query_cache.get user: @user_2).to.be.an.editable_version_of @user_2

      it 'returns the added entries', ->
        expect(@result.length).to.equal 2
        expect(@result[0]).to.be.an.editable_version_of @user_1
        expect(@result[1]).to.be.an.editable_version_of @user_2


    describe 'partial/full behavior', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @query_cache = new slender_data.QueryCache @test_params
        @query_cache.add_many [@partial_user_1, @full_user_2]
        @query_cache.add_many_as_clones [@full_user_1, @partial_user_2]

      it 'replaces an existing partial entry with the given full entry', ->
        expect(@query_cache.get username: 'one').to.eql @full_user_1

      it 'does not replace an existing full entry with the given partial entry', ->
        expect(@query_cache.get username: 'two').to.eql @full_user_2



  describe 'get', ->

    describe 'normal behavior', ->

      beforeEach ->
        @query_cache.add @user_1

      it 'returns the entry described by the given entry', ->
        expect(@query_cache.get user: @user_1).to.equal @user_1

      it 'returns the entry described by the given id', ->
        expect(@query_cache.get username: @user_1.username).to.equal @user_1

      it 'returns undefined if the entry does not exist', ->
        expect(@query_cache.get id: 99).to.be.undefined


    describe 'partial entries enabled', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @query_cache = new slender_data.QueryCache @test_params

      context 'partial entry cached', ->

        beforeEach ->
          @query_cache.add @partial_user_1

        it 'returns the partial entry if a partial entry is requested', ->
          @result = @query_cache.get username: 'one'
          expect(@result).to.eql @partial_user_1

        it 'returns nothing if a full entry is requested', ->
          @result = @query_cache.get username: 'one', full: yes
          expect(@result).to.be.undefined

      context 'full entry cached', ->

        beforeEach ->
          @query_cache.add @full_user_1

        it 'returns the full entry if a partial entry is requested', ->
          @result = @query_cache.get username: 'one'
          expect(@result).to.eql @full_user_1

        it 'returns the full entry if a full entry is requested', ->
          @result = @query_cache.get username: 'one', full: yes
          expect(@result).to.eql @full_user_1


    describe 'versioning', ->

      beforeEach ->
        @test_params.versioning = yes
        @query_cache = new slender_data.QueryCache @test_params

      it 'returns the cached entry if it is fresh', ->
        @query_cache.add @user_1
        @result = @query_cache.get user: @user_1
        expect(@result).to.eql @user_1

      it 'returns nothing if the cached entry is stale', ->
        @result = @query_cache.add @user_1_outdated
        expect(@result).to.be.undefined


  describe 'get_many', ->

    describe 'normal behavior', ->

      beforeEach ->
        @query_cache.add_many [@user_1, @user_2]

      it 'returns the entries described by the given entry multi-query', ->
        result = @query_cache.get_many users: [@user_1, @user_2]
        expect(result.found).to.eql [@user_1, @user_2]
        expect(result.missing.empty()).to.be.true

      it 'returns the entries described by the given id multi-query', ->
        result = @query_cache.get_many usernames: ['one', 'two']
        expect(result.found).to.eql [@user_1, @user_2]
        expect(result.missing.empty()).to.be.true

      it "reports non-existing entries as a MultiQuery in the 'missing' result portion", ->
        result = @query_cache.get_many usernames: ['one', 99]
        expect(result.missing).to.be.an.instanceof slender_data.MultiQuery
        expect(result.missing).to.eql_query usernames: [99]


    describe 'with partial entries enabled', ->

      beforeEach ->
        @test_params.partial_entries = yes
        @query_cache = new slender_data.QueryCache @test_params
        @query_cache.add_many [@partial_user_1, @full_user_2]

      context 'partial entries are allowed', ->

        beforeEach ->
          @result = @query_cache.get_many usernames: ['one', 'two', 'three']

        it 'returns partial and full entries', ->
          expect(@result.found).to.eql [@partial_user_1, @full_user_2]

        it 'reports missing entries', ->
          expect(@result.missing).to.eql_query usernames: ['three']

        it 'configures the missing entries multi_query to allow partial entries as well', ->
          expect(@result.missing.full).to.be.falsy


      context 'full entries requested', ->

        beforeEach ->
          @result = @query_cache.get_many usernames: ['one', 'two'], full: yes

        it 'returns only full entries', ->
          expect(@result.found).to.eql [@full_user_2]

        it 'reports missing entries', ->
          expect(@result.missing.to_hash().usernames).to.eql ['one']

        it 'configures the missing entries multi-query to also request only full entries', ->
          expect(@result.missing.full).to.be.true


    describe 'versioning', ->

      beforeEach ->
        @test_params.versioning = yes
        @query_cache = new slender_data.QueryCache @test_params
        @query_cache.add_many [@user_1_outdated, @user_2]
        @result = @query_cache.get_many users: [@user_1, @user_2]

      it 'returns only cached entries with the right version from the cache', ->
        expect(@result.found).to.eql [@user_2]

      it 'reports outdated cached entries as missing', ->
        expect(@result.missing).to.eql_query users: [@user_1]



  describe 'length', ->

    it 'returns the number of stored entries', ->
      expect(@query_cache.length()).to.equal 0
      @query_cache.add @user_1
      expect(@query_cache.length()).to.equal 1
      @query_cache.add @user_2
      expect(@query_cache.length()).to.equal 2
      @query_cache.remove user: @user_1
      expect(@query_cache.length()).to.equal 1


  describe 'remove', ->

    beforeEach ->
      @query_cache.add @user_1

    it 'removes the given entry from the cache', ->
      @query_cache.remove user: @user_1
      expect(@query_cache.length()).to.equal 0

    it 'removes the entry with the given id from the cache', ->
      @query_cache.remove username: 'one'
      expect(@query_cache.length()).to.equal 0


  describe 'remove_many', ->

    beforeEach ->
      @query_cache.add_many [@user_1, @user_2]

    it 'removes the given entries from the server', ->
      @query_cache.remove_many users: [@user_1, @user_2]
      expect(@query_cache.length()).to.equal 0

    it 'removes the entries with the given ids from the server', ->
      @query_cache.remove_many usernames: ['one', 'two']
      expect(@query_cache.length()).to.equal 0

